/* -*- Mode: C++; tab-width: 2; indent-tabs-mode: nil; c-basic-offset: 2 -*- */
/* vim:set ts=2 sw=2 sts=2 et cindent: */
/* ***** BEGIN LICENSE BLOCK *****
 * Version: MPL 1.1/GPL 2.0/LGPL 2.1
 *
 * The contents of this file are subject to the Mozilla Public License Version
 * 1.1 (the "License"); you may not use this file except in compliance with
 * the License. You may obtain a copy of the License at
 * http://www.mozilla.org/MPL/
 *
 * Software distributed under the License is distributed on an "AS IS" basis,
 * WITHOUT WARRANTY OF ANY KIND, either express or implied. See the License
 * for the specific language governing rights and limitations under the
 * License.
 *
 * The Original Code is Mozilla code.
 *
 * The Initial Developer of the Original Code is the Mozilla Corporation.
 * Portions created by the Initial Developer are Copyright (C) 2007
 * the Initial Developer. All Rights Reserved.
 *
 * Contributor(s):
 *  Chris Double <chris.double@double.co.nz>
 *  Chris Pearce <chris@pearce.org.nz>
 *
 * Alternatively, the contents of this file may be used under the terms of
 * either the GNU General Public License Version 2 or later (the "GPL"), or
 * the GNU Lesser General Public License Version 2.1 or later (the "LGPL"),
 * in which case the provisions of the GPL or the LGPL are applicable instead
 * of those above. If you wish to allow use of your version of this file only
 * under the terms of either the GPL or the LGPL, and not to allow others to
 * use your version of this file under the terms of the MPL, indicate your
 * decision by deleting the provisions above and replace them with the notice
 * and other provisions required by the GPL or the LGPL. If you do not delete
 * the provisions above, a recipient may use your version of this file under
 * the terms of any one of the MPL, the GPL or the LGPL.
 *
 * ***** END LICENSE BLOCK ***** */
#include "nsError.h"
#include "nsBuiltinDecoderStateMachine.h"
#include "nsBuiltinDecoder.h"
#include "nsMediaStream.h"
#include "nsWebMReader.h"
#include "VideoUtils.h"

using namespace mozilla;

// Un-comment to enable logging of seek bisections.
//#define SEEK_LOGGING

#ifdef PR_LOGGING
extern PRLogModuleInfo* gBuiltinDecoderLog;
#define LOG(type, msg) PR_LOG(gBuiltinDecoderLog, type, msg)
#ifdef SEEK_LOGGING
#define SEEK_LOG(type, msg) PR_LOG(gBuiltinDecoderLog, type, msg)
#else
#define SEEK_LOG(type, msg)
#endif
#else
#define LOG(type, msg)
#define SEEK_LOG(type, msg)
#endif

// TODO: Handle's only two streams - one audio and one video. Need's fixing/testing
//       for case of multiple streams.

// Chunk size to read when reading Ogg files. Average Ogg page length
// is about 4300 bytes, so we read the file in chunks larger than that.
static const int PAGE_STEP = 8192;

// Functions for reading and seeking using nsMediaStream required for
// nestegg_io. The 'user data' passed to these functions is the
// decoder from which the media stream is obtained.
static int webm_read(void *aBuffer, size_t aLength, void *aUserData)
{
  char *p = static_cast<char *>(aBuffer);
  nsBuiltinDecoder* decoder = reinterpret_cast<nsBuiltinDecoder*>(aUserData);
  nsMediaStream* stream = decoder->GetCurrentStream();
  nsresult rv = NS_OK;
  PRBool eof = PR_FALSE;
  while (NS_SUCCEEDED(rv) && aLength > 0) {
    PRUint32 bytes = 0;
    rv = stream->Read(p, aLength, &bytes);
    aLength -= bytes;
    p += bytes;
    if (bytes == 0) {
      eof = PR_TRUE;
      break;
    }
  }

  return NS_FAILED(rv) ? -1 : eof ? 0 : 1;
}

static int webm_seek(int64_t aOffset, int aWhence, void *aUserData)
{
  nsBuiltinDecoder* decoder = reinterpret_cast<nsBuiltinDecoder*>(aUserData);
  nsMediaStream* stream = decoder->GetCurrentStream();
  nsresult rv = stream->Seek(aWhence, aOffset);
  return NS_SUCCEEDED(rv) ? 0 : -1;
}

static int64_t webm_tell(void *aUserData)
{
  nsBuiltinDecoder* decoder = reinterpret_cast<nsBuiltinDecoder*>(aUserData);
  nsMediaStream* stream = decoder->GetCurrentStream();
  return stream->Tell();
}

nsWebMReader::nsWebMReader(nsBuiltinDecoder* aDecoder)
  : nsBuiltinDecoderReader(aDecoder),
  mContext(nsnull),
  mPacketCount(0),
  mChannels(0),
  mHasVideo(PR_FALSE),
  mHasAudio(PR_FALSE)
{
  MOZ_COUNT_CTOR(nsWebMReader);
}

nsWebMReader::~nsWebMReader()
{
  if (mContext) {
    nestegg_destroy(mContext);
  }

  mVideoPackets.Reset();
  mAudioPackets.Reset();

  vorbis_block_clear(&mVorbisBlock);
  vorbis_dsp_clear(&mVorbisDsp);
  vorbis_info_clear(&mVorbisInfo);
  vorbis_comment_clear(&mVorbisComment);

  MOZ_COUNT_DTOR(nsWebMReader);
}

nsresult nsWebMReader::Init() {
  if(vpx_codec_dec_init(&mVP8, &vpx_codec_vp8_dx_algo, NULL, 0)) {
    return NS_ERROR_FAILURE;
  }

  vorbis_info_init(&mVorbisInfo);
  vorbis_comment_init(&mVorbisComment);
  memset(&mVorbisDsp, 0, sizeof(vorbis_dsp_state));
  memset(&mVorbisBlock, 0, sizeof(vorbis_block));

  return NS_OK;
}

nsresult nsWebMReader::ResetDecode()
{
  nsresult res = NS_OK;
  if (NS_FAILED(nsBuiltinDecoderReader::ResetDecode())) {
    res = NS_ERROR_FAILURE;
  }

  // TODO: Do I need to reset vorbis state here?
  // TODO: Do I need to reset VP8 decoder state here?
  // TODO: Do I need to reset libnestegg state here?
  mVideoPackets.Reset();
  mAudioPackets.Reset();

  return res;
}

// TODO: need to cleanup packet data when returning due to errors!
nsresult nsWebMReader::ReadMetadata()
{
  NS_ASSERTION(mDecoder->OnStateMachineThread(), "Should be on play state machine thread.");
  MonitorAutoEnter mon(mMonitor);

  nestegg_io io;
  io.read = webm_read;
  io.seek = webm_seek;
  io.tell = webm_tell;
  io.userdata = static_cast<nsBuiltinDecoder*>(mDecoder);
  int r = nestegg_init(&mContext, io, NULL);
  if (r == -1) {
    return NS_ERROR_FAILURE;
  }

  uint64_t duration = 0;
  r = nestegg_duration(mContext, &duration);
  if (r == -1) {
    return NS_ERROR_FAILURE;
  }
  mDecoder->SetDuration(duration / 1000000);
  printf("Duration: %ld\n", (long)(duration / 1000000));

  unsigned int ntracks = 0;
  r = nestegg_track_count(mContext, &ntracks);
  if (r == -1) {
    return NS_ERROR_FAILURE;
  }
  printf("Tracks: %d\n", ntracks);

  mInfo.mHasAudio = PR_FALSE;
  mInfo.mHasVideo = PR_FALSE;
  for (PRUint32 i = 0; i < ntracks; ++i) {
    int id = nestegg_track_codec_id(mContext, i);
    if (id == -1) {
      return NS_ERROR_FAILURE;
    }
    int type = nestegg_track_type(mContext, i);
    printf("Track %d codec id: %d type %d ", i, id, type);
    if (type == NESTEGG_TRACK_VIDEO) {
      nestegg_video_params params;
      r = nestegg_track_video_params(mContext, i, &params);
      if (r == -1) {
        return NS_ERROR_FAILURE;
      }
      printf("%dx%d\n", params.width, params.height);

      mHasVideo = PR_TRUE;
      mInfo.mHasVideo = PR_TRUE;
      mInfo.mPicture.x = 0;
      mInfo.mPicture.y = 0;
      mInfo.mPicture.width = params.width;
      mInfo.mPicture.height = params.height;
      mInfo.mFrame.width = params.width;
      mInfo.mFrame.height = params.height;

      // TODO: What do I set this too?
      mInfo.mDataOffset = -1;

      mInfo.mPixelAspectRatio = float(params.display_width) / params.width;

      // TODO: add framerate when nestgets support
      mInfo.mFramerate = 25.0;
      mInfo.mCallbackPeriod = 1000 / mInfo.mFramerate;
    }
    if (type == NESTEGG_TRACK_AUDIO) {
      nestegg_audio_params params;
      r = nestegg_track_audio_params(mContext, i, &params);
      if (r == -1) {
        return NS_ERROR_FAILURE;
      }
      printf(" rate: %f channels: %d depth: %d", (float)params.rate, params.channels, params.depth);

      mHasAudio = PR_TRUE;
      mInfo.mHasAudio = PR_TRUE;
      mInfo.mAudioRate = params.rate;
      mInfo.mAudioChannels = params.channels;
      mChannels = params.channels;

      if (!mInfo.mHasVideo) {
        mInfo.mCallbackPeriod = 1.0 / 25.0;
      }

      // Get the Vorbis header data
      // TODO: Factor this out somehow
      unsigned int nheaders = 0;
      r = nestegg_track_codec_data_count(mContext, i, &nheaders);
      if (r == -1 || nheaders != 3) {
        return NS_ERROR_FAILURE;
      }

      for (PRUint32 header = 0; header < nheaders; ++header) {
        ogg_packet opacket;
        unsigned char* data = 0;
        size_t length = 0;

        r = nestegg_track_codec_data(mContext, i, header, &data, &length);
        // TODO: uncomment error checking when nestegg bug is fixed
#if 0
        if (r == -1) {
          printf("nestegg_track_codec_data failed\n");
          return NS_ERROR_FAILURE;
        }
#endif
        opacket.packet = data;
        opacket.bytes = length;
        opacket.e_o_s = 0;
        opacket.b_o_s = header == 0;
        opacket.granulepos = 0;
        opacket.packetno = mPacketCount++;
     
        r = vorbis_synthesis_headerin(&mVorbisInfo,
                                      &mVorbisComment,
                                      &opacket);
        if (r < 0) {
          return NS_ERROR_FAILURE;
        }
      }

      r = vorbis_synthesis_init(&mVorbisDsp, &mVorbisInfo);
      if (r < 0) {
        return NS_ERROR_FAILURE;
      }

      r = vorbis_block_init(&mVorbisDsp, &mVorbisBlock);
      if (r < 0) {
        return NS_ERROR_FAILURE;
      }
    }
    printf("\n");
  }

  return NS_OK;
}

// TODO: cleanup nestegg_packet on error
PRBool nsWebMReader::DecodeAudioPacket(nestegg_packet* aPacket)
{
  mMonitor.AssertCurrentThreadIn();

  int r = 0;
  unsigned int count = 0;
  r = nestegg_packet_count(aPacket, &count);
  if (r == -1) {
    return PR_FALSE;
  }

  for (PRUint32 i = 0; i < count; ++i) {
    unsigned char* data;
    size_t length;
    r = nestegg_packet_data(aPacket, i, &data, &length);
    if (r == -1) {
      return PR_FALSE;
    }

    uint64_t tstamp = 0;
    r = nestegg_packet_tstamp(aPacket, &tstamp);
    if (r == -1) {
      return PR_FALSE;
    }

    ogg_packet opacket;
    opacket.packet = data;
    opacket.bytes = length;
    opacket.e_o_s = 0;
    opacket.b_o_s = 0;
    opacket.granulepos = -1;
    opacket.packetno = mPacketCount++;

    if (vorbis_synthesis(&mVorbisBlock, &opacket) != 0) {
      return PR_FALSE;
    }

    if (vorbis_synthesis_blockin(&mVorbisDsp,
                                 &mVorbisBlock) != 0) {
      return PR_FALSE;
    }

    float** pcm = 0;
    PRUint32 samples = 0;
    while ((samples = vorbis_synthesis_pcmout(&mVorbisDsp, &pcm)) > 0) {
      if (samples > 0) {
        float* buffer = new float[samples * mChannels];
        float* p = buffer;
        for (PRUint32 i = 0; i < samples; ++i) {
          for (PRUint32 j = 0; j < mChannels; ++j) {
            *p++ = pcm[j][i];
          }
        }

        PRInt64 duration = samples * 1000 / mVorbisDsp.vi->rate;
        SoundData* s = new SoundData(0, //mPageOffset,
                                     tstamp / 1000000,
                                     duration,
                                     samples,
                                     buffer,
                                     mChannels);
        mAudioQueue.Push(s);
      }
      if (vorbis_synthesis_read(&mVorbisDsp, samples) != 0) {
        return PR_FALSE;
      }
    }
  }

  nestegg_free_packet(aPacket);

  return PR_TRUE;
}

// TODO: Need to decode audio only during DecodeAudioData and push video packets
// onto a queue for later processing.
PRBool nsWebMReader::DecodeAudioData()
{
  MonitorAutoEnter mon(mMonitor);
  NS_ASSERTION(mDecoder->OnStateMachineThread() || mDecoder->OnDecodeThread(),
               "Should be on playback or decode thread.");
  int r = 0;
  nestegg_packet* packet = 0;

  if (mAudioPackets.GetSize() > 0) {
    packet = mAudioPackets.PopFront();
  }
  else {
    r = nestegg_read_packet(mContext, &packet);
    if (r == 0) {
      // End of stream
      mVideoQueue.Finish();
      mAudioQueue.Finish();
      return PR_FALSE;
    }

    if (r < 0) {
      // TODO: What to do when there is an error?
      printf("Error: %d\n", r);
      return PR_FALSE;
    }
  }

  unsigned int track = 0;
  r = nestegg_packet_track(packet, &track);
  if (r == -1) {
    return PR_FALSE;
  }

  if (nestegg_track_type(mContext, track) != NESTEGG_TRACK_AUDIO) {
    mVideoPackets.Push(packet);
    return PR_TRUE;
  }

  return DecodeAudioPacket(packet);
}

// TODO: Need to decode video only during DecodeVideoFrame and push audio packets
// onto a queue for later processing.
PRBool nsWebMReader::DecodeVideoFrame(PRBool &aKeyframeSkip,
                                     PRInt64 aTimeThreshold)
{
  MonitorAutoEnter mon(mMonitor);
  NS_ASSERTION(mDecoder->OnStateMachineThread() || mDecoder->OnDecodeThread(),
               "Should be on state machine or AV thread.");
  int r = 0;
  nestegg_packet* packet = 0;
  // Check if we have queued video packets already and get one of those if so
  if (mVideoPackets.GetSize() > 0) {
    packet = mVideoPackets.PopFront();
  }
  else {
    r = nestegg_read_packet(mContext, &packet);
    if (r == 0) {
      // End of stream
      mVideoQueue.Finish();
      mAudioQueue.Finish();
      return PR_FALSE;
    }

    if (r < 0) {
      // TODO: What to do when there is an error?
      printf("Error: %d\n", r);
      return PR_FALSE;
    }
  }

  unsigned int track = 0;
  r = nestegg_packet_track(packet, &track);
  if (r == -1) {
    return PR_FALSE;
  }

  if (nestegg_track_type(mContext, track) != NESTEGG_TRACK_VIDEO) {
    mAudioPackets.Push(packet);
    return PR_TRUE;
  }

  unsigned int count = 0;
  r = nestegg_packet_count(packet, &count);
  if (r == -1) {
    return PR_FALSE;
  }

  for (PRUint32 i = 0; i < count; ++i) {
    unsigned char* data;
    size_t length;
    r = nestegg_packet_data(packet, i, &data, &length);
    if (r == -1) {
      return PR_FALSE;
    }

    uint64_t tstamp = 0;
    r = nestegg_packet_tstamp(packet, &tstamp);
    if (r == -1) {
      return PR_FALSE;
    }

//    printf("DecodeVideoFrame(%s,%lld) tstamp: %lld\n", aKeyframeSkip ? "skip": "noskip", aTimeThreshold, tstamp/1000000);

    vpx_codec_stream_info_t si;
    si.sz = sizeof(si);
    si.w = si.h = 0;
    si.is_kf = 0;
    if (vpx_codec_peek_stream_info(&vpx_codec_vp8_dx_algo, data, length, &si) == 0) {
//      printf("vpx_codec_peek_steam_info(%d %dx%d %s)\n", si.sz, si.w, si.h, si.is_kf ? "key" : "notkey");
    }

    if ((aKeyframeSkip && !si.is_kf) || (aKeyframeSkip && si.is_kf && tstamp/1000000 < static_cast<uint64_t>(aTimeThreshold))) {
//      printf("%lld %lld - skip\n", tstamp/1000000, aTimeThreshold);
      aKeyframeSkip = PR_TRUE;
      break;
    }

    if (aKeyframeSkip && si.is_kf) {
//      printf("%lld %lld - finished skipping\n", tstamp/1000000, aTimeThreshold);
      aKeyframeSkip = PR_FALSE;
    }

    if(vpx_codec_decode(&mVP8, data, length, NULL, 0)) {
      return PR_FALSE;
    }

    // If the timestamp of the video frame is less than
    // the time threshold required then it is not added
    // to the video queue and won't be displayed.
    if (tstamp / 1000000 < static_cast<uint64_t>(aTimeThreshold)) {
      continue;
    }

    vpx_codec_iter_t  iter = NULL;
    vpx_image_t      *img;

    while((img = vpx_codec_get_frame(&mVP8, &iter))) {
      VideoData::YCbCrBuffer b;
      b.mPlanes[0].mData = img->planes[0];
      b.mPlanes[0].mStride = img->stride[0];
      b.mPlanes[0].mHeight = img->d_h;
      b.mPlanes[0].mWidth = img->d_w;

      b.mPlanes[1].mData = img->planes[1];
      b.mPlanes[1].mStride = img->stride[1];
      b.mPlanes[1].mHeight = img->d_h >> 1;
      b.mPlanes[1].mWidth = img->d_w >> 1;
 
      b.mPlanes[2].mData = img->planes[2];
      b.mPlanes[2].mStride = img->stride[2];
      b.mPlanes[2].mHeight = img->d_h >> 1;
      b.mPlanes[2].mWidth = img->d_w >> 1;
  
      VideoData *v = VideoData::Create(mInfo,
                                       mDecoder->GetImageContainer(),
                                       0, /* mPageOffset */
                                       tstamp / 1000000,
                                       b,
                                       PR_FALSE, /* isKeyframe */
                                       tstamp / 1000000);
      mVideoQueue.Push(v);
    }
  }
 
  nestegg_free_packet(packet);
 
  return PR_TRUE;
}

PRInt64 nsWebMReader::FindEndTime(PRInt64 aEndOffset)
{
  MonitorAutoEnter mon(mMonitor);
  NS_ASSERTION(mDecoder->OnStateMachineThread(), "Should be on state machine thread.");
  return 0;
}

nsresult nsWebMReader::Seek(PRInt64 aTarget, PRInt64 aStartTime, PRInt64 aEndTime)
{
  MonitorAutoEnter mon(mMonitor);
  NS_ASSERTION(mDecoder->OnStateMachineThread(),
               "Should be on state machine thread.");
  LOG(PR_LOG_DEBUG, ("%p About to seek to %lldms", mDecoder, aTarget));
  if (NS_FAILED(ResetDecode())) {
    return NS_ERROR_FAILURE;
  }
  int r = nestegg_track_seek(mContext, 0, aTarget * 1000000);
  if (r != 0) {
    return NS_ERROR_FAILURE;
  }
  if (HasVideo()) {
    nsAutoPtr<VideoData> video;
    PRBool eof = PR_FALSE;
    PRInt64 startTime = -1;
    video = nsnull;
    while (HasVideo() && !eof) {
      while (mVideoQueue.GetSize() == 0 && !eof) {
        PRBool skip = PR_FALSE;
        eof = !DecodeVideoFrame(skip, 0);
        MonitorAutoExit exitReaderMon(mMonitor);
        MonitorAutoEnter decoderMon(mDecoder->GetMonitor());
        if (mDecoder->GetDecodeState() == nsBuiltinDecoderStateMachine::DECODER_STATE_SHUTDOWN) {
          return NS_ERROR_FAILURE;
        }
      }
      if (mVideoQueue.GetSize() == 0) {
        break;
      }
      video = mVideoQueue.PeekFront();
      // If the frame end time is less than the seek target, we won't want
      // to display this frame after the seek, so discard it.
      if (video && video->mTime + 40 < aTarget) {
        if (startTime == -1) {
          startTime = video->mTime;
        }
        mVideoQueue.PopFront();
        video = nsnull;
      } else {
        video.forget();
        break;
      }
    }
    SEEK_LOG(PR_LOG_DEBUG, ("First video frame after decode is %lld", startTime));
  }
  return NS_OK;
}

